from dataclasses import is_dataclass
from datetime import datetime
from typing import get_type_hints, Any
from typing import get_origin, get_args, Union, Callable, Mapping, Iterable, Generator
from collections.abc import Iterable as ABCIterable, Mapping as ABCMapping, Generator as ABCGenerator

from sqlmodel import SQLModel

from pyonir.utilities import get_attr


def is_iterable(tp):
    origin = get_origin(tp) or tp
    return isinstance(origin, type) and issubclass(origin, ABCIterable) and not issubclass(origin, (str, bytes))

def is_generator(tp):
    origin = get_origin(tp) or tp
    return isinstance(origin, type) and issubclass(origin, ABCGenerator)

def is_mappable_type(tp):
    if tp == dict: return True
    origin = get_origin(tp)
    args = get_args(tp)
    return isinstance(origin, type) and issubclass(origin, ABCMapping)

def is_scalar_type(tp):
    return tp in (int, float, str, bool)

def is_custom_class(tp):
    return isinstance(tp, type) and not tp.__module__ == "builtins"

def unwrap_optional(tp):
    """Unwrap Optional[T] → T, else return tp unchanged"""
    origin_tp = get_origin(tp)
    if is_mappable_type(origin_tp):
        key_tp, value_tp = get_args(tp)
        return origin_tp, key_tp, unwrap_optional(value_tp)
    if is_iterable(origin_tp):
        value_tps = get_args(tp)
        return origin_tp, value_tps
    if origin_tp is Union:
        args = [unwrap_optional(a) for a in get_args(tp) if a is not type(None)]
        if len(args):
            res = [arg for arg, *rest in args]
            return res
    return tp, None

def is_callable_type(tp):
    return get_origin(tp) is Callable

def is_optional_type(tp):
    return get_origin(tp) is Union and type(None) in get_args(tp)

def is_option_type(t):
    if get_origin(t) is not Union: return t
    return [arg for arg in get_args(t) if arg is not type(None)][0]

def coerce_union(t, v):
    try:
        return t(v)
    except Exception as exc:
        print(f"failed to coerce {v} into {t}")
        return None

def collect_type_hints(t):
    hints = get_type_hints(t)
    try:
        init_hints = get_type_hints(t.__init__)
        hints.update(init_hints)
        del hints['return']
    except Exception as exc:
        pass
    return hints

# def add_props_to_object(res, data: dict, file_src: object =None, attrs: list = None):
#     """Add all properties from data and file_src to res object"""
#     itr = [*data.keys(), *(attrs or [])]
#     for key in itr:
#         if key[0] == '_': continue
#         # if frozen and not hasattr(res, key): continue
#         # if isinstance(val, property): continue
#         value = get_attr(data, key) or get_attr(file_src, key)
#         setattr(res, key, value)
#     return res

def required_parameters(cls):
    import inspect
    sig = inspect.signature(cls.__init__)
    required = []
    for name, param in sig.parameters.items():
        if name in ("self","args","kwargs"):  # skip self, *args, **kwargs
            continue
        if param.default is inspect.Parameter.empty:
            required.append(name)
    return required

def set_attr(target: object, attr: str, value: Any):
    if isinstance(target, dict):
        target.update({attr: value})
    else:
        setattr(target, attr, value)

def func_request_mapper(func: Callable, data: object) -> dict: pass

def cls_mapper(file_obj: object, cls: Union[type, list[type]], from_request=None):
    """Recursively map dict-like input into `cls` with type-safe field mapping."""
    from pyonir.core import PyonirRequest, PyonirApp

    if hasattr(cls, '__skip_parsely_deserialization__'):
        return file_obj

    # Union types
    if isinstance(cls, (list, tuple, set)):
        _value = None
        for ct in cls:
            if isinstance(file_obj, ct):
                _value = ct(file_obj)
            if _value is not None: break
            _value = coerce_union(ct, file_obj)
        return _value

    # datetime passthrough
    if cls == datetime:
        return file_obj

    # Scalars just wrap
    if is_scalar_type(cls):
        return cls(file_obj)
    is_sqlmodel = lambda t: isinstance(t, type) and issubclass(t, SQLModel)
    is_dclass = is_dataclass(cls) or len(required_parameters(cls)) > 0
    is_generic_type = cls.__name__ == 'GenericQueryModel'
    param_type_map =  {k: field.annotation for k, field in cls.model_fields.items()} if is_sqlmodel(cls) else collect_type_hints(cls)
    data = get_attr(file_obj, 'data') or {}

    # Merge nested access if ORM opts define mapper_key
    orm_opts = getattr(cls, "_orm_options", {})
    mapper_keys = orm_opts.get("mapper", {})
    is_frozen = orm_opts.get("frozen")
    access_path_to_nested = '.'.join(['data', orm_opts.get('mapper_key', cls.__name__.lower())])
    nested_value = get_attr(file_obj, access_path_to_nested)
    if nested_value:
        data.update(**nested_value)

    cls_args = {} if from_request or is_dclass else cls()
    for name, hint in param_type_map.items():
        mapper_name = get_attr(mapper_keys, name, None) #or name
        if name.startswith("_") or name == "return":
            continue

        actual_type, *mapable = unwrap_optional(hint)
        # value = None
        for ds in (data, file_obj, cls): # try to get value from data, file_obj, cls (in that order)
            value = get_attr(ds, mapper_name or name)
            if value is not None: break

        if value is None:
            set_attr(cls_args, name, value)
            continue
        # Handle Special Pyonir objects
        if from_request and hint in (PyonirApp, PyonirRequest):
            from pyonir import Site
            value = Site if hint == PyonirApp else from_request
            set_attr(cls_args, name, value)
            continue
        if from_request and hint == value:
            set_attr(cls_args, name, None)
            continue

        # Handle containers
        custom_mapper_fn = getattr(cls, f'map_to_{name}', None)
        if custom_mapper_fn:
            value = custom_mapper_fn(value)
        elif is_sqlmodel(hint):
            value = cls_mapper(value, hint) if isinstance(value, dict) else value
        elif is_scalar_type(actual_type):
            value = actual_type(value)
        elif is_callable_type(actual_type) or callable(value):
            pass
        elif is_mappable_type(actual_type) and len(mapable) and mapable[0]:
            key_type, value_types = mapable
            vtype = value_types[0] if len(value_types)==1 else None
            value = {key_type(k): cls_mapper(v, vtype or value_types) for k, v in value.items()}
        elif is_iterable(actual_type):
            itypes = mapable[0] if len(mapable) == 1 else mapable
            itype = itypes[0] if itypes and len(itypes) == 1 else None
            value = [cls_mapper(v, itype) for v in value] if itype else value if isinstance(value, actual_type) else actual_type(value)
        elif isinstance(value, actual_type):
            pass
        elif is_custom_class(actual_type):
            value = cls_mapper(value, actual_type)
        else:
            try:
                value = actual_type(value)
            except Exception as e:
                pass
        set_attr(cls_args, name, value)

    # Methods passed from request are returned to be called later
    if callable(cls) and from_request:
        return cls_args

    res = cls(**cls_args) if is_dclass else cls_args
    if is_generic_type:
        data = file_obj if not data and isinstance(file_obj, dict) else data
        keys = list(cls.__dict__.keys()) + ['file_name','file_created_on']
        for key in keys:
            if key[0] == '_': continue
            _key = mapper_keys.get(key, key)
            value = get_attr(data, _key or key) or get_attr(file_obj, _key or key)
            set_attr(res, key, value)
        return res

    # Pass additional fields that are not specified on model
    if not is_frozen and not is_sqlmodel(cls):
        # keys = data.keys() if is_sqlmodel(cls) else data.model_fields.keys()
        for key, value in data.items():
            if isinstance(getattr(cls, key, None), property):
                continue  # skip properties
            if param_type_map.get(key) or key[0] == '_':
                continue  # skip private or declared attributes
            setattr(res, key, value)

    return res

def dict_to_class(data: dict, name: Union[str, callable] = None, deep: bool = True) -> object:
    """
    Converts a dictionary into a class object with the given name.

    Args:
        data (dict): The dictionary to convert.
        name (str): The name of the class.
        deep (bool): If True, convert all dictionaries recursively.
    Returns:
        object: An instance of the dynamically created class with attributes from the dictionary.
    """
    # Dynamically create a new class
    cls = type(name or 'T', (object,), {}) if not callable(name) and deep!='update' else name

    # Create an instance of the class
    instance = cls() if deep!='update' else cls
    setattr(instance, 'update', lambda d: dict_to_class(d, instance, 'update') )
    # Assign dictionary keys as attributes of the instance
    for key, value in data.items():
        if isinstance(getattr(cls, key, None), property): continue
        if deep and isinstance(value, dict):
            value = dict_to_class(value, key)
        setattr(instance, key, value)

    return instance



import os
from dataclasses import dataclass
from typing import Optional, Dict

from jinja2 import Environment
from pyonir.models.app import BaseApp


class TemplateEnvironment(Environment):

    def __init__(self, app: BaseApp):
        if not os.path.exists(app.frontend_dirpath) and app.use_themes:
            raise ValueError(f"Frontend directory {app.frontend_dirpath} does not exist. Please ensure the frontend directory is set up correctly.")
        from jinja2 import FileSystemLoader, ChoiceLoader
        from pyonir import PYONIR_JINJA_TEMPLATES_DIRPATH, PYONIR_JINJA_FILTERS_DIRPATH, PYONIR_JINJA_EXTS_DIRPATH
        from webassets.ext.jinja2 import AssetsExtension
        from pyonir.utilities import load_modules_from

        installed_extensions = load_modules_from(PYONIR_JINJA_EXTS_DIRPATH, True)
        app_extensions = [AssetsExtension, *installed_extensions]
        jinja_template_paths = ChoiceLoader([FileSystemLoader(app.frontend_templates_dirpath), FileSystemLoader(PYONIR_JINJA_TEMPLATES_DIRPATH)])
        super().__init__(loader=jinja_template_paths, extensions=app_extensions)

        #  Custom filters
        sys_filters = load_modules_from(PYONIR_JINJA_FILTERS_DIRPATH)
        app_filters = load_modules_from(app.template_filters_dirpath)
        app_filters = {**sys_filters, **app_filters}
        self.filters.update(**app_filters)

        def url_for(path):
            rmaps = app.server.url_map if app.server else {}
            return rmaps.get(path, {}).get('path', '/'+path)

        # Include globals
        self.globals['url_for'] = url_for
        self.globals['request'] = None
        self.globals['user'] = None


    def load_template_path(self, template_path: str):
        """Adds template path to file loader"""
        from jinja2 import FileSystemLoader
        app_loader = self.loader
        if not app_loader: return
        self.loader.loaders.append(FileSystemLoader(template_path))

    def add_filter(self, filter: callable):
        name = filter.__name__
        print(f"Installing filter:{name}")
        self.filters.update({name: filter})
        pass

@dataclass
class Theme:
    _orm_options = {'mapper': {'theme_dirname': 'file_dirname', 'theme_dirpath': 'file_dirpath'}}
    name: str
    theme_dirname: str = ''
    """Directory name for theme folder within frontend/themes directory"""
    theme_dirpath: str = ''
    """Directory path for theme folder within frontend/themes directory"""
    details: Optional['Parsely'] = None
    """Represents a theme available in the frontend/themes directory."""
    _template_dirname = 'templates'
    _static_dirname = 'static'

    def __post_init__(self):
        self.details = self.readme()
        for k, v in self.details.data.items():
            if k in ('static_dirname', 'templates_dirname'):
                setattr(self, k, v)

    @property
    def static_dirname(self):
        """directory name for theme's jinja templates"""
        return self.details.data.get('static_dirname', self._static_dirname) if self.details else self._static_dirname

    @property
    def templates_dirname(self):
        """directory name for theme's jinja templates"""
        return self.details.data.get('templates_dirname', self._template_dirname) if self.details else self._template_dirname

    @property
    def static_dirpath(self):
        """directory to serve static theme assets"""
        return os.path.join(self.theme_dirpath, self.static_dirname)

    @property
    def jinja_template_path(self):
        return os.path.join(self.theme_dirpath, self.templates_dirname)

    def readme(self):
        """Returns the theme's README.md file content if available"""
        from pyonir import Site
        from pyonir.models.parser import DeserializeFile
        theme_ctx = list(Site.app_ctx)
        theme_ctx[2] = Site.frontend_dirpath
        theme_readme = os.path.join(self.theme_dirpath,'README.md')
        theme_readme =  theme_readme if os.path.exists(theme_readme) else os.path.join(self.theme_dirpath,'readme.md')
        readme = DeserializeFile(theme_readme, app_ctx=theme_ctx)
        if not readme.file_exists:
            raise ValueError(f"Theme {self.name} does not have a README.md file.")
        return readme

class PyonirThemes:
    """Represents sites available and active theme(s) within the frontend directory."""

    def __init__(self, theme_dirpath: str):
        if not os.path.exists(theme_dirpath):
            raise ValueError(f"Theme directory {theme_dirpath} does not exist.")
        self.themes_dirpath: str = theme_dirpath # directory path to available site themes
        self.available_themes: Optional[Dict[str, Theme]] = self.query_themes() # collection of themes available in frontend/themes directory

    @property
    def active_theme(self) -> Optional[Theme]:
        from pyonir import Site
        from pyonir.models.utils import get_attr
        if not Site or not self.available_themes: return None
        site_theme = get_attr(Site.settings, 'app.theme_name')
        site_theme = self.available_themes.get(site_theme)
        return site_theme

    def query_themes(self) -> Optional[Dict[str, Theme]]:
        """Returns a collection of available themes within the frontend/themes directory"""
        from pyonir import Site
        themes_map = {}
        for theme_dir in os.listdir(self.themes_dirpath):
            if theme_dir.startswith(BaseApp.IGNORE_WITH_PREFIXES): continue
            theme = Theme(name=theme_dir, theme_dirname=theme_dir,
                          theme_dirpath=os.path.join(self.themes_dirpath, theme_dir))
            theme._template_dirname = Site.TEMPLATES_DIRNAME
            theme._static_dirname = Site.FRONTEND_ASSETS_DIRNAME
            themes_map[theme_dir] = theme
        return themes_map if themes_map else None




import os, json
import textwrap

from pyonir.models.parser import DeserializeFile
from pyonir.tests.backend.demo_controller import DemoService
from pyonir import init

def generate_dataclass_from_class(cls, output_dir="types"):
    from typing import get_type_hints
    attr_map = get_type_hints(cls)
    props_map = {k: type(v).__name__ for k, v in cls.__dict__.items() if isinstance(v, property)}
    meth_map = {k: callable for k, v in cls.__dict__.items() if callable(v)}
    all_map = dict(**props_map, **meth_map, **attr_map)
    lines = [f"class {cls.__name__}:"]
    if not cls.__annotations__:
        lines.append("    pass")
    else:
        for name, typ in all_map.items():
            lines.append(f"    {name}: {typ.__class__.__name__}")
    with open(os.path.join(os.path.dirname(__file__), output_dir, f"{cls.__name__}.py"), "w") as f:
        f.write("\n".join(lines))

def generate_py_tests(parsely: DeserializeFile):
    from pyonir.models.utils import create_file
    cases = []
    name = parsely.__class__.__name__
    # indent = " " * 4
    test_setup = textwrap.dedent(
        """        
        import pytest, os
        from pyonir.models.parser import DeserializeFile
        true = True
        false = False
        parselyFile = DeserializeFile(os.path.join(os.path.dirname(__file__),'contents', 'test.md'))\n\n
        # test cases for {name}
        {test_cases}
        """
    )
    test_case = """
    def test_{case_name}():
        obj = {value}
        assert obj == parselyFile.data.get('{case_name}')
    """

    for case_name, value in parsely.data.items():
        _case = test_case.format(case_name=case_name, value=json.dumps(value))
        cases.append(textwrap.dedent(_case))

    final_test_case = test_setup.format(test_cases="\n".join(cases), name=name)
    create_file(os.path.join(os.path.dirname(__file__), f'test_{name.lower()}.py'), final_test_case)

if __name__=='__main__':
    app_dirpath = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'libs', 'app_setup')
    App = init(os.path.join(app_dirpath,'main.py'), use_themes=False)
    file = DeserializeFile(os.path.join(os.path.dirname(__file__),'contents','test.md'))
    # generate_parsely_tests(file)
    generate_py_tests(file)

    pass


import os
from datetime import datetime
from collections.abc import Generator
from typing import Optional, Union

from pyonir.models.mapper import cls_mapper


def process_contents(path, app_ctx=None, file_model: any = None) -> object:
    """Deserializes all files within the contents directory"""
    from pyonir.models.database import query_fs
    key = os.path.basename(path)
    res = type(key, (object,), {"__name__": key})() # generic map
    pgs = query_fs(path, app_ctx=app_ctx, model=file_model)
    for pg in pgs:
        name = getattr(pg, 'file_name')
        # pg_obj = type(name, (object,), {"__name__": name, 'file_path': pg.file_path})
        # val = cls_mapper(pg, pg_obj)
        setattr(res, name, pg.to_named_tuple())
    return res

def json_serial(obj):
    """JSON serializer for nested objects not serializable by default jsonify"""
    from datetime import datetime
    if isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, Generator) or hasattr(obj, 'mapping'):
        return list(obj)
    elif hasattr(obj, 'to_dict'):
        return obj.to_dict()

def deserialize_datestr(
    datestr: Union[str, datetime],
    fmt: str = "%Y-%m-%d %I:%M:%S",   # %I for 12-hour format
    zone: str = "US/Eastern",
    auto_correct: bool = True
) -> Optional[datetime]:
    """
    Convert a date string into a timezone-aware datetime.

    Args:
        datestr: Input string or datetime.
        fmt: Expected datetime format (default "%Y-%m-%d %I:%M:%S %p").
        zone: Timezone name (default "US/Eastern").
        auto_correct: Whether to attempt corrections for sloppy inputs.

    Returns:
        Timezone-aware datetime (in UTC), or None if parsing fails.
    """
    import pytz

    if isinstance(datestr, datetime):
        return pytz.utc.localize(datestr) if datestr.tzinfo is None else datestr.astimezone(pytz.utc)
    if not isinstance(datestr, str):
        return None

    tz = pytz.timezone(zone)

    def correct_format(raw: str, dfmt: str) -> tuple[str, str]:
        """Try to normalize sloppy date strings like 2025/8/9 13:00."""
        try:
            raw = raw.strip().replace("/", "-")
            if 'T' in raw:
                date_part, _, time_part = raw.partition('T')
            else:
                date_part, _, time_part = raw.partition(" ")

            # Use fallback timestr if missing
            time_part = time_part or "12:00:00.0000"
            hr,*minsec = time_part.split(':')
            is_military_tme = "%H" in dfmt or int(hr) > 12
            dfmt = dfmt.replace("%I", "%H") if is_military_tme else fmt

            parts = date_part.split("-")
            if len(parts) != 3:
                return raw, dfmt

            y, m, d = parts
            # Pad month/day
            m, d = f"{int(m):02d}", f"{int(d):02d}"

            # Basic sanity check: if year looks like day
            if int(y) < int(d):
                # Swap year/day (common human error)
                y, d = d, y
                print(f"⚠️  Corrected malformed date string: {raw} → {y}-{m}-{d}")

            return f"{y}-{m}-{d} {time_part}", dfmt
        except Exception as e:
            return raw, dfmt

    try:
        # Try direct parse first
        dt = datetime.strptime(datestr, fmt)
    except ValueError:
        if not auto_correct:
            return None
        corrected, fmt = correct_format(datestr, fmt)
        if not corrected:
            return None
        try:
            dt = datetime.strptime(corrected, fmt)
        except ValueError:
            return None

    # Localize to input zone, then convert to UTC
    return tz.localize(dt).astimezone(pytz.utc)

def get_attr(rowObj, attrPath=None, default=None, rtn_none=True):
    """
    Resolves nested attribute or dictionary key paths.

    Args:
        obj: the root object
        attr_path: dot-separated string or list for nested access
        default: fallback value if the target is None or missing
        return_none: if True, returns `None` on missing keys/attrs instead of the original object

    Returns:
        The nested value, or `default`, or `obj` based on fallback rules.
    """
    if attrPath == None: return rowObj
    attrPath = attrPath if isinstance(attrPath, list) else attrPath.split('.')
    targetObj = None
    for key in attrPath:
        try:
            if targetObj:
                targetObj = targetObj[key]
            else:
                targetObj = rowObj.get(key)
            pass
        except (KeyError, AttributeError, TypeError) as e:
            if targetObj:
                targetObj = getattr(targetObj, key, None)
            else:
                targetObj = getattr(rowObj, key, None)
            pass
    if targetObj is None and rtn_none:
        return default or None

    return targetObj


def create_file(file_abspath: str, data: any = None, is_json: bool = False, mode='w') -> bool:
    """Creates a new file based on provided data
    Args:
        file_abspath: str = path to proposed file
        data: any = contents to write into file
        is_json: bool = strict json file
        mode: str = write mode for file w|w+|a
    Returns:
        bool: The return value if file was created successfully
    """
    def write_file(file_abspath, data, is_json=False, mode='w'):
        import json
        with open(file_abspath, mode, encoding="utf-8") as f:
            if is_json:
                json.dump(data, f, indent=2, sort_keys=True, default=json_serial)
            else:
                f.write(data)

    if not os.path.exists(os.path.dirname(file_abspath)):
        os.makedirs(os.path.dirname(file_abspath))
    try:
        is_json = is_json or file_abspath.endswith('.json')
        write_file(file_abspath, data, is_json=is_json, mode=mode)
        return True
    except Exception as e:
        print(f"Error create_file method: {str(e)}")
        return False

def copy_assets(src: str, dst: str, purge: bool = True):
    """Copies files from a source directory into a destination directory with option to purge destination"""
    import shutil
    from shutil import ignore_patterns
    # print(f"{PrntColrs.OKBLUE}Coping `{src}` resource into {dst}{PrntColrs.RESET}")
    try:
        if os.path.exists(dst) and purge:
            shutil.rmtree(dst)
        if os.path.isfile(src):
            shutil.copyfile(src, dst)
        if os.path.isdir(src):
            shutil.copytree(src, dst, ignore=ignore_patterns('__pycache__', '*.pyc', 'tmp*', 'node_modules', '.*'))
    except Exception as e:
        raise




def asjson(input, escaped = False):
    """converts input parameter into a json string. pyonir json_serial is used to convert non supported data types"""
    from pyonir.utilities import json_serial
    import json, html
    d = json.dumps(input, default=json_serial)
    if escaped: return html.escape(d)
    return d
    # return d.replace('<script>','<\/script>').replace('</script>','<\/script>')
